context("Sanity checks")

test_that("required arguments are provided", {
  expect_error(
    Apply(), 
    "missing, with no default"
  )
  expect_error(
    Apply(1:10), 
    "missing, with no default"
  )
  expect_error(
    Apply(1:10, fun = mean), 
    "must be specified"
  )
})

test_that("arguments have the right type", {
  expect_equal(
    Apply(1:10, NULL, mean),
    list(output1 = array(1:10, dim = 10))
  )
  expect_error(
    Apply(numeric(0), NULL, mean),
    "must be of length > 0"
  )
  #expect_error(
  #)
})


#context("Parameter order")
#
#test_that("", {
#  expect_that(
#    Apply(),
#    throws_error("")
#  )
#})


context("Use cases")

test_that("in1: 1 val; targ. dims: 0; out1: 0 val", {
  f <- function(x) NULL
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(logical(0), dim = c(0, 1)))
  )
  f <- function(x) numeric(0)
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(logical(0), dim = c(0, 1)))
  )
})

test_that("in1: 1 val; targ. dims: 0; out1: 0 val", {
  f <- function(x) NULL
  expect_equal(
    Apply(1, 1, f), 
    list(output1 = array(logical(0)))
  )
  f <- function(x) numeric(0)
  expect_equal(
    Apply(1, 1, f), 
    list(output1 = array(logical(0)))
  )
})

test_that("in1: 1 val; targ. dims: 0; out1: 1 val", {
  expect_equal(
    Apply(1, NULL, mean), 
    list(output1 = array(1))
  )
})

test_that("in1: 1 val; targ. dims: 1; out1: 1 val", {
  expect_equal(
    Apply(1, 1, mean), 
    list(output1 = 1)
  )
})

test_that("in1: 1 val; targ. dims: 0; out1: 3 val", {
  f <- function(x) x:(x + 2)
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(1:3, dim = c(3, 1)))
  )
})

test_that("in1: 1 val; targ. dims: 1; out1: 3 val", {
  f <- function(x) x:(x + 2)
  expect_equal(
    Apply(1, 1, f), 
    list(output1 = array(1:3))
  )
})

test_that("in1: 1 val; targ. dims: 0; out1: 1 dim", {
  # unnamed output dim
  f <- function(x) array(x:(x + 2))
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(1:3, dim = c(3, 1)))
  )
  # named output dim
  f <- function(x) array(x:(x + 2), dim = c(time = 3))
  expect_equal(
    Apply(1, NULL, f),
    list(output1 = array(1:3, dim = c(time = 3, 1)))
  )
})

test_that("in1: 1 val; targ. dims: 1; out1: 1 dim", {
  # unnamed output dim
  f <- function(x) array(x:(x + 2))
  expect_equal(
    Apply(1, 1, f), 
    list(output1 = array(1:3))
  )
  # named output dim
  f <- function(x) array(x:(x + 2), dim = c(time = 3))
  expect_equal(
    Apply(1, 1, f), 
    list(output1 = array(1:3, dim = c(time = 3)))
  )
})

test_that("in1: 1 val; targ. dims: 0; out1: 2 dim", {
  # unnamed output dims
  f <- function(x) array(x:(x + 2), dim = c(3, 4))
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(1:3, dim = c(3, 4, 1)))
  )
  # named output dim 1
  f <- function(x) array(x:(x + 2), dim = c(time = 3, 4))
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(1:3, dim = c(time = 3, 4, 1)))
  )
  # named output dim 2
  f <- function(x) array(x:(x + 2), dim = c(3, region = 4))
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(1:3, dim = c(3, region = 4, 1)))
  )
  # named output dims
  f <- function(x) array(x:(x + 2), dim = c(time = 3, region = 4))
  expect_equal(
    Apply(1, NULL, f), 
    list(output1 = array(1:3, dim = c(time = 3, region = 4, 1)))
  )
})

test_that("in1: 1 val; targ. dims: 1; out1: 2 dim", {
  # unnamed output dims
  f <- function(x) array(x:(x + 2), dim = c(3, 4))
  expect_equal(
    Apply(1, 1, f), 
    list(output1 = array(1:3, dim = c(3, 4)))
  )
  # named output dims
  f <- function(x) array(x:(x + 2), dim = c(time = 3, region = 4))
  expect_equal(
    Apply(1, 1, f), 
    list(output1 = array(1:3, dim = c(time = 3, region = 4)))
  )
})

test_that("in1: 1 val; targ. dims: 0; out1: 1 val; out 2: 1 val", {
  # unnamed outputs
  f <- function(x) list(mean(x), mean(x) + 1)
  expect_equal(
    Apply(1, NULL, f),
    list(output1 = array(1),
         output2 = array(2))
  )
  # named outputs
  f <- function(x) list(a = mean(x), b = mean(x) + 1)
  expect_equal(
    Apply(1, NULL, f),
    list(a = array(1),
         b = array(2))
  )
})

test_that("in1: 1 val; targ. dims: 1; out1: 3 val; out 2: 2 dim", {
  f <- function(x) list(a = x:(x + 2), array(x:(x + 2), dim = c(3, region = 4)))
  expect_equal(
    Apply(1, 1, f),
    list(a = array(1:3),
         output2 = array(1:3, dim = c(3, region = 4)))
  )
})

test_that("in1: 1 dim; targ. dims: 0; out1: 0 val", {
  # unnamed dim
  # unnamed output
  f <- function(x) NULL
  expect_equal(
    Apply(array(1:10), NULL, f), 
    list(output1 = array(logical(0), dim = c(0, 10)))
  )
  # named dim
  # named output
  f <- function(x) list(out1 = NULL)
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), NULL, f), 
    list(out1 = array(logical(0), dim = c(0, a = 10)))
  )
})

test_that("in1: 1 dim; targ. dims: 1; out1: 0 val", {
  # unnamed dim
  # unnamed output
  f <- function(x) NULL
  expect_equal(
    Apply(array(1:10), 1, f), 
    list(output1 = array(logical(0)))
  )
  # named dim
  # named output
  f <- function(x) list(out1 = NULL)
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 1, f), 
    list(out1 = array(logical(0), dim = c(0)))
  )
  # named target dim
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 'a', f), 
    list(out1 = array(logical(0), dim = c(0)))
  )
})

test_that("in1: 1 dim; targ. dims: 0; out1: 1 val", {
  # unnamed dim
  # unnamed output
  expect_equal(
    Apply(array(1:10), NULL, mean), 
    list(output1 = array(1:10))
  )
  # named dim
  # named output
  f <- function(x) list(out1 = mean(x))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), NULL, f), 
    list(out1 = array(1:10, dim = c(a = 10)))
  )
})

test_that("in1: 1 dim; targ. dims: 1; out1: 1 val", {
  # unnamed dim
  # unnamed output
  expect_equal(
    Apply(array(1:10), 1, mean), 
    list(output1 = 5.5)
  )
  # named dim
  # named output
  f <- function(x) list(out1 = mean(x))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 1, f), 
    list(out1 = 5.5)
  )
  # named target dim
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 'a', f), 
    list(out1 = 5.5)
  )
})

test_that("in1: 1 dim; targ. dims: 0; out1: 3 vals", {
  # unnamed dim
  # unnamed output
  f <- function(x) x:(x + 2)
  expect_equal(
    Apply(array(1:10), NULL, f), 
    list(output1 = array(sapply(1:10, function(x) x:(x + 2)), dim = c(3, 10)))
  )
  # named dim
  # named output
  f <- function(x) list(out1 = x:(x + 2))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), NULL, f), 
    list(out1 = array(sapply(1:10, function(x) x:(x + 2)), dim = c(3, a = 10)))
  )
})

test_that("in1: 1 dim; targ. dims: 1; out1: 3 vals", {
  # unnamed dim
  # unnamed output
  f <- function(x) x[1]:(x[1] + 2)
  expect_equal(
    Apply(array(1:10), 1, f), 
    list(output1 = array(1:3))
  )
  # named dim
  # named output
  f <- function(x) list(out1 = x[1]:(x[1] + 2))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 1, f), 
    list(out1 = array(1:3))
  )
  # named target dim
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 'a', f), 
    list(out1 = array(1:3))
  )
})

test_that("in1: 1 dim; targ. dims: 0; out1: 1 dim", {
  # unnamed input dim
  # unnamed output
  # unnamed output dim
  f <- function(x) array(x:(x + 2))
  expect_equal(
    Apply(array(1:10), NULL, f), 
    list(output1 = array(sapply(1:10, function(x) x:(x + 2)), dim = c(3, 10)))
  )
  # named input dim
  # named output
  # named output dim
  f <- function(x) list(out1 = array(x:(x + 2), dim = c(b = 3)))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), NULL, f), 
    list(out1 = array(sapply(1:10, function(x) x:(x + 2)), dim = c(b = 3, a = 10)))
  )
})

test_that("in1: 1 dim; targ. dims: 1; out1: 1 dim", {
  # unnamed input dim
  # unnamed output
  # unnamed output dim
  f <- function(x) array(x[1]:(x[1] + 2))
  expect_equal(
    Apply(array(1:10), 1, f), 
    list(output1 = array(1:3))
  )
  # named input dim
  # named output
  # named output dim
  f <- function(x) list(out1 = array(x[1]:(x[1] + 2), dim = c(b = 3)))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 1, f), 
    list(out1 = array(1:3, dim = c(b = 3)))
  )
  # named target dim
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 'a', f), 
    list(out1 = array(1:3, dim = c(b = 3)))
  )
})

test_that("in1: 1 dim; targ. dims: 0; out1: 2 dims", {
  # unnamed input dim
  # unnamed output
  # unnamed output dims
  f <- function(x) array(x:(x + 2), dim = c(3, 4))
  expect_equal(
    Apply(array(1:10), NULL, f), 
    list(output1 = array(sapply(1:10, function(x) rep(x:(x + 2), 4)), dim = c(3, 4, 10)))
  )
  # named input dim
  # named output
  # named output dim
  f <- function(x) list(out1 = array(x:(x + 2), dim = c(b = 3, c = 4)))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), NULL, f), 
    list(out1 = array(sapply(1:10, function(x) rep(x:(x + 2), 4)), dim = c(b = 3, c = 4, a = 10)))
  )
})

test_that("in1: 1 dim; targ. dims: 1; out1: 2 dims", {
  # unnamed input dim
  # unnamed output
  # unnamed output dim
  f <- function(x) array(x[1]:(x[1] + 2), dim = c(3, 4))
  expect_equal(
    Apply(array(1:10), 1, f), 
    list(output1 = array(sapply(1:10, function(x) rep(x:(x + 2), 4)), dim = c(3, 4)))
  )
  # named input dim
  # named output
  # named output dim
  f <- function(x) list(out1 = array(x[1]:(x[1] + 2), dim = c(b = 3, c = 4)))
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 1, f), 
    list(out1 = array(sapply(1:10, function(x) rep(x:(x + 2), 4)), dim = c(b = 3, c = 4)))
  )
  # named target dim
  expect_equal(
    Apply(array(1:10, dim = c(a = 10)), 'a', f), 
    list(out1 = array(sapply(1:10, function(x) rep(x:(x + 2), 4)), dim = c(b = 3, c = 4)))
  )
})

# FAILS
test_that("in1: 1 dim; targ. dims: 0; out1: 1 dim; out2: 2 dims; out3: 1 val", {
  f <- function(x) list(array(x:(x + 2), dim = c(3)),
                        array(x:(x + 3), dim = c(4, 5)),
                        mean(x))
  expect_equal(
    Apply(array(1:10), NULL, f),
    list(output1 = array(sapply(1:10, function(x) x:(x + 2)), dim = c(3, 10)),
         output2 = array(sapply(1:10, function(x) rep(x:(x + 3), 5)), dim = c(4, 5, 10)),
         output3 = array(1:10))
  )
})

test_that("in1: 1 dim; targ. dims: 1; out1: 1 dim; out2: 2 dims; out3: 1 val", {
})

test_that("in1: 2 dim; targ. dims: 0; out1: 1 dim; out2: 2 dims; out3: 1 val", {
})

test_that("in1: 2 dim; targ. dims: 1; out1: 1 dim; out2: 2 dims; out3: 1 val", {
# first target dim
## without name
## with name
# second target dim
## without name
## with name
})

test_that("in1: 2 dim; targ. dims: 2; out1: 1 dim; out2: 2 dims; out3: 1 val", {
# without name
# with name
})

test_that("in1: 3 dim; targ. dims: 0; out1: 1 dim; out2: 2 dims; out3: 1 val", {
})

test_that("in1: 3 dim; targ. dims: 1; out1: 1 dim; out2: 2 dims; out3: 1 val", {
# first target dim
## without name
## with name
# second target dim
## without name
## with name
# third target dim
## without name
## with name
})

test_that("in1: 3 dim; targ. dims: 2; out1: 1 dim; out2: 2 dims; out3: 1 val", {
# first couple
## without name
## with name
# second couple
## without name
## with name
# extremes couple
## without name
## with name
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 0; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 0; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 1; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 1; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 0; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 0; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 1; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 1; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 0; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 0; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 1; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 1; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 0; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 0; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 1; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 1; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 0; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 0; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 0, 1; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 1 val; in2: 1 val; targ. dims: 1, 1; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})










test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 0; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 0; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 1; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 1; out1: 0 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 0; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 0; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 1; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 1; out1: 1 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 0; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 0; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 1; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 1; out1: 3 val", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 0; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 0; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 1; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 1; out1: 1 dim", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 0; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 0; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 0, 1; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 1 val; in2: 1 dim; targ. dims: 1, 1; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})










test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 0; out1: 0 val", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 0; out1: 0 val", {
#first in first dim
#first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 1; out1: 0 val", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 1; out1: 0 val", {
#shared target dims
##first in first dim
##first in second dim
#not shared target dims
##first in first dim
##first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 2, 1; out1: 0 val", {
# shared first target dim
# shared second target dim
# not shared target dims
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 0; out1: 1 val", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 0; out1: 1 val", {
#first in first dim
#first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 1; out1: 1 val", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 1; out1: 1 val", {
#shared target dims
##first in first dim
##first in second dim
#not shared target dims
##first in first dim
##first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 2, 1; out1: 1 val", {
# shared first target dim
# shared second target dim
# not shared target dims
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 0; out1: 3 val", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 0; out1: 3 val", {
#first in first dim
#first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 1; out1: 3 val", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 1; out1: 3 val", {
#shared target dims
##first in first dim
##first in second dim
#not shared target dims
##first in first dim
##first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 2, 1; out1: 3 val", {
# shared first target dim
# shared second target dim
# not shared target dims
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 0; out1: 1 dim", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 0; out1: 1 dim", {
#first in first dim
#first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 1; out1: 1 dim", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 1; out1: 1 dim", {
#shared target dims
##first in first dim
##first in second dim
#not shared target dims
##first in first dim
##first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 2, 1; out1: 1 dim", {
# shared first target dim
# shared second target dim
# not shared target dims
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 0; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 0; out1: 1 dim; out2: 1 val; out3: 3 dim", {
#first in first dim
#first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 0, 1; out1: 1 dim; out2: 1 val; out3: 3 dim", {
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 1, 1; out1: 1 dim; out2: 1 val; out3: 3 dim", {
#shared target dims
##first in first dim
##first in second dim
#not shared target dims
##first in first dim
##first in second dim
})

test_that("in1: 2 dim; in2: 1 dim; targ. dims: 2, 1; out1: 1 dim; out2: 1 val; out3: 3 dim", {
# shared first target dim
# shared second target dim
# not shared target dims
})












test_that("in1: 2 dim; in2: 3 dim; targ. dims: 0-2, 0-3; out1: 2 dim", {
# shared first target dim
# shared second target dim
# shared two target dims (first two in second in)
# shared two target dims (last two in second in)
# shared two target dims (extreme two in second in)
# not shared target dims
})

test_that("in1: 2 dim; in2: 3 dim; targ. dims: 0-2, 0-3; out1: 1 dim; out2: 1 val; out3: 3 dim", {
# shared first target dim
# shared second target dim
# shared two target dims (first two in second in)
# shared two target dims (last two in second in)
# shared two target dims (extreme two in second in)
# not shared target dims
})

test_that("in1: 2 dim; in2: 3 dim; in3: 1 dim; targ. dims: 0-2, 0-3, 0-1; out1: 2 dim", {
# shared first target dim
# shared second target dim
# shared two target dims (first two in second in)
# shared two target dims (last two in second in)
# shared two target dims (extreme two in second in)
# not shared target dims
})



# TODOS:
# TESTS FOR MARGINS
# TESTS FOR DISORDERED TARGET_DIMS
# TESTS FOR FUN WITH TARGET_DIMS AND OUTPUT_DIMS ATTACHED
# TESTS FOR FUNCTIONS RECEIVING ADDITIONAL PARAMETERS
# TESTS FOR SPLIT FACTOR
# TESTS FOR NCORES
# TESTS OF WALLCLOCK TIME
# TESTS OF MEMORY FOOTPRINT
